float getSSAO(in vec3 screenPos, in vec3 viewNormal) {
    // Dithering for anti-aliasing based on the frame's randomness
    vec3 dither = (ANTI_ALIASING >= 2) ? fract(getRng3(ivec2(gl_FragCoord.xy) & 255) + frameFract) : getRng3(ivec2(gl_FragCoord.xy) & 255);
    
    // Initialize occlusion value
    float occlusion = 0.25;

    // Base dither offset for sampling
    vec3 baseDither = (dither.xyz - 0.5) * 0.5;
    
    // Predefined offsets for SSAO sampling
    vec3 ditherOffsets[4] = vec3[4](
        baseDither,
        baseDither.zxy,
        baseDither.yzx,
        (fract(dither.zyx + GOLDEN_RATIO) - 0.5) * 0.5
    );

    // Calculate depth origin from screen position
    float depthOrigin = near / (1.0 - screenPos.z);
    
    // Calculate base position in world space
    vec3 basePos = getViewPos(gbufferProjectionInverse, screenPos) + viewNormal * 0.5;

    // Sample surrounding positions for occlusion
    for (int i = 0; i < 4; i++) {
        // Calculate sample position by adding dither offset to the base position
        vec3 samplePos = getScreenPos(gbufferProjection, basePos + ditherOffsets[i]);
        
        // Sample depth from the depth texture and linearize it
        float sampleDepth = textureLod(depthtex0, samplePos.xy, 0).x;

        // Check if the sampled position is occluded
        if (samplePos.z > sampleDepth) {
            // Calculate occlusion contribution based on depth difference
            float depthDiff = depthOrigin - near / (1.0 - sampleDepth);
            occlusion -= 0.0625 / max(depthDiff, 1.0);
        }
    }

    // Remap occlusion value to a reasonable range [0, 1]
    return clamp(occlusion, 0.0, 1.0);
}
