vec3 basicShadingForward(in vec4 albedo) {
    // Initialize total diffuse with the sky color
    vec3 totalDiffuse = toLinear(SKY_COLOR_DATA_BLOCK);

    #ifdef IS_IRIS
        // Add lightning flash effect if applicable
        totalDiffuse += lightningFlash;
    #endif

    #ifndef CLOUDS
        // Calculate sky light squared to modulate diffuse lighting
        float skyLightSquared = squared(lmCoord.y);
        totalDiffuse *= skyLightSquared; // Occlude by sky light

        // Add block light contribution
        totalDiffuse += toLinear(lmCoord.x * blockLightColor);
    #endif

    // Add ambient lighting, considering night vision
    totalDiffuse += toLinear(nightVision * 0.5 + AMBIENT_LIGHTING);

    #ifdef WORLD_LIGHT
        #ifdef SHADOW_MAPPING
            // Transform vertex position for shadow mapping
            vec3 shdPos = vec3(vertexShdPos.xy / (length(vertexShdPos.xy) * 2.0 + 0.2) + 0.5, vertexShdPos.z);

            // Apply bias for shadow mapping if necessary
            #ifdef CLOUDS
                const vec3 biasAdjustMult = vec3(2, 2, -0.0625) * shadowMapPixelSize;
                shdPos += vec3(vertexNLX, vertexNLY, vertexNLZ) * biasAdjustMult; // Normal-based bias
            #endif

            // Sample shadows with optional filtering
            vec3 shdCol;
            #ifdef SHADOW_FILTER
                float blueNoise = (ANTI_ALIASING >= 2) 
                    ? fract(texelFetch(noisetex, ivec2(gl_FragCoord.xy) & 255, 0).x + frameFract)
                    : texelFetch(noisetex, ivec2(gl_FragCoord.xy) & 255, 0).x;

                shdCol = getShdCol(shdPos, blueNoise * TAU);
            #else
                shdCol = getShdCol(shdPos);
            #endif

            // Adjust shadow color based on fade factor
            float shdFactor = shdFade;

            #ifdef CLOUDS
                shdFactor *= max(0.0, vertexNLZ * 0.6 + 0.4); // Cloud influence on shadowing
            #endif

            shdCol *= shdFactor; // Apply shadow factor to shadow color
        #else
            // Handle cases without shadow mapping
            float shdCol;
            #ifdef CLOUDS
                shdCol = max(0.0, vertexNLZ * 0.6 + 0.4) * shdFade; // Simple diffuse for clouds
            #else
                shdCol = saturate(hermiteMix(0.96, 0.98, lmCoord.y)) * shdFade; // Fake shadows
            #endif
        #endif

        #ifndef FORCE_DISABLE_WEATHER
            // Approximate rain's effect on light diffusion
            float rainDiffuseAmount = rainStrength * 0.5;
            shdCol *= (1.0 - rainDiffuseAmount); // Reduce shadow color by rain effect

            #ifdef CLOUDS
                shdCol += rainDiffuseAmount; // Add rain effect for clouds
            #else
                shdCol += rainDiffuseAmount * skyLightSquared; // Add rain effect modulated by sky light
            #endif
        #endif

        // Add shadow diffuse contribution to total diffuse color
        totalDiffuse += shdCol * toLinear(LIGHT_COLOR_DATA_BLOCK0);
    #endif

    // Return final shaded color by multiplying with albedo
    return albedo.rgb * totalDiffuse;
}
